"""
Show materialized configuration of tox environments.
"""

from textwrap import indent
from typing import Dict, Iterable, List, Set

from colorama import Fore

from tox.config.cli.parser import ToxParser
from tox.config.loader.stringify import stringify
from tox.config.sets import ConfigSet
from tox.plugin import impl
from tox.session.cmd.run.common import env_run_create_flags
from tox.session.common import CliEnv, env_list_flag
from tox.session.state import State
from tox.tox_env.api import ToxEnv
from tox.tox_env.errors import Skip
from tox.tox_env.package import PackageToxEnv
from tox.tox_env.runner import RunToxEnv


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("config", ["c"], "show tox configuration", show_config)
    our.add_argument(
        "-k", nargs="+", help="list just configuration keys specified", dest="list_keys_only", default=[], metavar="key"
    )
    our.add_argument(
        "--core", action="store_true", help="show core options too when selecting an env with -e", dest="show_core"
    )
    env_list_flag(our, default=CliEnv("ALL"))
    env_run_create_flags(our, mode="config")


def show_config(state: State) -> int:
    is_colored = state.options.is_colored
    keys: List[str] = state.options.list_keys_only
    is_first = True

    def _print_env(tox_env: ToxEnv) -> None:
        nonlocal is_first
        if is_first:
            is_first = False
        else:
            print("")
        print_section_header(is_colored, f"[testenv:{tox_env.conf.name}]")
        if not keys:
            print_key_value(is_colored, "type", type(tox_env).__name__)
        print_conf(is_colored, tox_env.conf, keys)

    def _get_run_env(env_name: str) -> RunToxEnv:
        try:
            return state.tox_env(env_name)
        except Skip:
            return state.tox_env(env_name)  # get again to get the temporary state

    # because the target env could be a packaging one we first need to discover all defined ones
    run_envs: Dict[str, RunToxEnv] = {}
    pkg_envs: Dict[str, PackageToxEnv] = {}
    for name in state.env_list(everything=True):
        run_env = _get_run_env(name)
        run_envs[name] = run_env
        for pkg_env in run_env.package_envs:
            pkg_envs[pkg_env.conf.name] = pkg_env

    show_everything = state.options.env.all
    done_pkg_envs: Set[str] = set()
    for name in state.env_list():  # now go through selected ones
        if name in pkg_envs:
            if name not in done_pkg_envs:
                _print_env(pkg_envs[name])
                done_pkg_envs.add(name)
        else:
            run_env = run_envs[name] if name in run_envs else _get_run_env(name)  # an on-demand env, construct it now
            _print_env(run_env)
            if show_everything:
                for pkg_env in run_env.package_envs:
                    if pkg_env.name not in done_pkg_envs:
                        _print_env(pkg_env)
                        done_pkg_envs.add(pkg_env.name)

    # environments may define core configuration flags, so we must exhaust first the environments to tell the core part
    if show_everything or state.options.show_core:
        print("")
        print_section_header(is_colored, "[tox]")
        print_conf(is_colored, state.conf.core, keys)
    return 0


def _colored(is_colored: bool, color: int, msg: str) -> str:
    return f"{color}{msg}{Fore.RESET}" if is_colored else msg


def print_section_header(is_colored: bool, name: str) -> None:
    print(_colored(is_colored, Fore.YELLOW, name))


def print_comment(is_colored: bool, comment: str) -> None:
    print(_colored(is_colored, Fore.CYAN, comment))


def print_key_value(is_colored: bool, key: str, value: str, multi_line: bool = False) -> None:
    print(_colored(is_colored, Fore.GREEN, key), end="")
    print(" =", end="")
    if multi_line:
        print("")
        value_str = indent(value, prefix="  ")
    else:
        print(" ", end="")
        value_str = value
    print(value_str)


def print_conf(is_colored: bool, conf: ConfigSet, keys: Iterable[str]) -> None:
    for key in keys if keys else conf:
        if key not in conf:
            continue
        key = conf.primary_key(key)
        try:
            value = conf[key]
            as_str, multi_line = stringify(value)
        except Exception as exception:  # because e.g. the interpreter cannot be found
            as_str, multi_line = _colored(is_colored, Fore.LIGHTRED_EX, f"# Exception: {exception!r}"), False
        if multi_line and "\n" not in as_str:
            multi_line = False
        print_key_value(is_colored, key, as_str, multi_line=multi_line)
    unused = conf.unused()
    if unused and not keys:
        print_comment(is_colored, f"# !!! unused: {', '.join(unused)}")



import inspect
import re
from concurrent.futures import Future
from configparser import ConfigParser, SectionProxy
from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator, List, Optional, Set, Type, TypeVar

from tox.config.loader.api import Loader, Override
from tox.config.loader.ini.factor import filter_for_env
from tox.config.loader.ini.replace import replace
from tox.config.loader.str_convert import StrConvert
from tox.config.set_env import SetEnv
from tox.report import HandledError

if TYPE_CHECKING:
    from tox.config.main import Config

V = TypeVar("V")
_COMMENTS = re.compile(r"(\s)*(?<!\\)#.*")


class IniLoader(StrConvert, Loader[str]):
    """Load configuration from an ini section (ini file is a string to string dictionary)"""

    def __init__(self, section: str, parser: ConfigParser, overrides: List[Override], core_prefix: str) -> None:
        self._section: SectionProxy = parser[section]
        self._parser = parser
        self.core_prefix = core_prefix
        super().__init__(overrides)

    def load_raw(self, key: str, conf: Optional["Config"], env_name: Optional[str]) -> str:
        return self.process_raw(conf, env_name, self._section[key])

    @staticmethod
    def process_raw(conf: Optional["Config"], env_name: Optional[str], value: str) -> str:
        # strip comments
        elements: List[str] = []
        for line in value.split("\n"):
            if not line.startswith("#"):
                part = _COMMENTS.sub("", line)
                elements.append(part.replace("\\#", "#"))
        strip_comments = "\n".join(elements)
        if conf is None:  # conf is None when we're loading the global tox configuration file for the CLI
            factor_filtered = strip_comments  # we don't support factor and replace functionality there
        else:
            factor_filtered = filter_for_env(strip_comments, env_name)  # select matching factors
        collapsed = factor_filtered.replace("\r", "").replace("\\\n", "")  # collapse explicit new-line escape
        return collapsed

    @contextmanager
    def build(
        self,
        future: "Future[V]",
        key: str,
        of_type: Type[V],
        conf: Optional["Config"],
        env_name: Optional[str],
        raw: str,
        chain: List[str],
    ) -> Generator[str, None, None]:
        delay_replace = inspect.isclass(of_type) and issubclass(of_type, SetEnv)

        def replacer(raw_: str, chain_: List[str]) -> str:
            if conf is None:
                replaced = raw_  # no replacement supported in the core section
            else:
                try:
                    replaced = replace(conf, env_name, self, raw_, chain_)  # do replacements
                except Exception as exception:
                    if isinstance(exception, HandledError):
                        raise
                    msg = f"replace failed in {'tox' if env_name is None else env_name}.{key} with {exception!r}"
                    raise HandledError(msg) from exception
            return replaced

        if not delay_replace:
            raw = replacer(raw, chain)
        yield raw
        if delay_replace:
            converted = future.result()
            if hasattr(converted, "replacer"):  # pragma: no branch
                converted.replacer = replacer  # type: ignore[attr-defined]

    def found_keys(self) -> Set[str]:
        return set(self._section.keys())

    def get_section(self, name: str) -> Optional[SectionProxy]:
        # needed for non tox environment replacements
        if self._parser.has_section(name):
            return self._parser[name]
        return None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(section={self._section}, overrides={self.overrides!r})"



"""
Run tox environments in parallel.
"""
import logging
from argparse import ArgumentParser, ArgumentTypeError
from typing import Optional

from tox.config.cli.parser import ToxParser
from tox.plugin import impl
from tox.session.common import env_list_flag
from tox.session.state import State
from tox.util.cpu import auto_detect_cpus

from .common import env_run_create_flags, execute

logger = logging.getLogger(__name__)

ENV_VAR_KEY = "TOX_PARALLEL_ENV"
OFF_VALUE = 0
DEFAULT_PARALLEL = OFF_VALUE


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("run-parallel", ["p"], "run environments in parallel", run_parallel)
    env_list_flag(our)
    env_run_create_flags(our, mode="run-parallel")
    parallel_flags(our, default_parallel=auto_detect_cpus())


def parse_num_processes(str_value: str) -> Optional[int]:
    if str_value == "all":
        return None
    if str_value == "auto":
        return auto_detect_cpus()
    try:
        value = int(str_value)
    except ValueError as exc:
        raise ArgumentTypeError(f"value must be a positive number, is {str_value!r}") from exc
    if value < 0:
        raise ArgumentTypeError(f"value must be positive, is {value!r}")
    return value


def parallel_flags(our: ArgumentParser, default_parallel: int) -> None:
    our.add_argument(
        "-p",
        "--parallel",
        dest="parallel",
        help="run tox environments in parallel, the argument controls limit: all,"
        " auto - cpu count, some positive number, zero is turn off",
        action="store",
        type=parse_num_processes,
        default=default_parallel,
        metavar="VAL",
    )
    our.add_argument(
        "-o",
        "--parallel-live",
        action="store_true",
        dest="parallel_live",
        help="connect to stdout while running environments",
    )
    our.add_argument(
        "--parallel-no-spinner",
        action="store_true",
        dest="parallel_no_spinner",
        help="do not show the spinner",
    )


def run_parallel(state: State) -> int:
    """here we'll just start parallel sub-processes"""
    return execute(
        state,
        max_workers=state.options.parallel,
        has_spinner=state.options.parallel_no_spinner is False and state.options.parallel_live is False,
        live=state.options.parallel_live,
    )



"""
Apply value substitution (replacement) on tox strings.
"""
import os
import re
import sys
from configparser import SectionProxy
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

from tox.config.loader.stringify import stringify
from tox.config.set_env import SetEnv
from tox.config.sets import ConfigSet
from tox.execute.request import shell_cmd

if TYPE_CHECKING:
    from tox.config.loader.ini import IniLoader
    from tox.config.main import Config

BASE_TEST_ENV = "testenv"

# split alongside :, unless it's escaped, or it's preceded by a single capital letter (Windows drive letter in paths)
ARGS_GROUP = re.compile(r"(?<!\\\\|:[A-Z]):")


def replace(conf: "Config", name: Optional[str], loader: "IniLoader", value: str, chain: List[str]) -> str:
    # perform all non-escaped replaces
    end = 0
    while True:
        start, end, to_replace = find_replace_part(value, end)
        if to_replace is None:
            break
        replaced = _replace_match(conf, name, loader, to_replace, chain.copy())
        if replaced is None:
            # if we cannot replace, keep what was there, and continue looking for additional replaces following
            # note, here we cannot raise because the content may be a factorial expression, and in those case we don't
            # want to enforce escaping curly braces, e.g. it should work to write: env_list = {py39,py38}-{,dep}
            end = end + 1
            continue
        new_value = value[:start] + replaced + value[end + 1 :]
        end = 0  # if we performed a replacement start over
        if new_value == value:  # if we're not making progress stop (circular reference?)
            break
        value = new_value
    # remove escape sequences
    value = value.replace("\\{", "{")
    value = value.replace("\\}", "}")
    value = value.replace("\\[", "[")
    value = value.replace("\\]", "]")
    return value


REPLACE_PART = re.compile(
    r"""
        (?<! \\) \{  # Unescaped {
            ( [^{}] | \\ \{ | \\ \} )*  # Anything except an unescaped { or }
        (?<! \\) \}  # Unescaped }
    |
        (?<! \\) \[ \]  # Unescaped []
    """,
    re.VERBOSE,
)  # simplified - not verbose version (?<!\\)([^{}]|\\\{|\\\})*(?<!\\)\}|(?<!\\)\[\]


def find_replace_part(value: str, end: int) -> Tuple[int, int, Optional[str]]:
    match = REPLACE_PART.search(value, end)
    if match is None:
        return -1, -1, None
    if match.group() == "[]":
        return match.start(), match.end() - 1, "posargs"  # brackets is an alias for positional arguments
    return match.start(), match.end() - 1, match.group()[1:-1]


def _replace_match(
    conf: "Config", current_env: Optional[str], loader: "IniLoader", value: str, chain: List[str]
) -> Optional[str]:
    of_type, *args = ARGS_GROUP.split(value)
    if of_type == "/":
        replace_value: Optional[str] = os.sep
    elif of_type == "" and args == [""]:
        replace_value = os.pathsep
    elif of_type == "env":
        replace_value = replace_env(conf, current_env, args, chain)
    elif of_type == "tty":
        replace_value = replace_tty(args)
    elif of_type == "posargs":
        replace_value = replace_pos_args(conf, current_env, args)
    else:
        replace_value = replace_reference(conf, current_env, loader, value, chain)
    return replace_value


_REPLACE_REF = re.compile(
    rf"""
    (\[(?P<full_env>{BASE_TEST_ENV}(:(?P<env>[^]]+))?|(?P<section>[-\w]+))\])? # env/section
    (?P<key>[a-zA-Z0-9_]+) # key
    (:(?P<default>.*))? # default value
""",
    re.VERBOSE,
)


def replace_reference(
    conf: "Config",
    current_env: Optional[str],
    loader: "IniLoader",
    value: str,
    chain: List[str],
) -> Optional[str]:
    # a return value of None indicates could not replace
    match = _REPLACE_REF.match(value)
    if match:
        settings = match.groupdict()

        key = settings["key"]
        if settings["section"] is None and settings["full_env"]:
            settings["section"] = settings["full_env"]

        exception: Optional[Exception] = None
        try:
            for src in _config_value_sources(settings["env"], settings["section"], current_env, conf, loader):
                try:
                    if isinstance(src, SectionProxy):
                        return loader.process_raw(conf, current_env, src[key])
                    value = src.load(key, chain)
                    as_str, _ = stringify(value)
                    as_str = as_str.replace("#", r"\#")  # escape comment characters as these will be stripped
                    return as_str
                except KeyError as exc:  # if fails, keep trying maybe another source can satisfy
                    exception = exc
        except Exception as exc:
            exception = exc
        if exception is not None:
            if isinstance(exception, KeyError):  # if the lookup failed replace - else keep
                default = settings["default"]
                if default is not None:
                    return default
                # we cannot raise here as that would mean users could not write factorials: depends = {py39,py38}-{,b}
            else:
                raise exception
    return None


def _config_value_sources(
    env: Optional[str],
    section: Optional[str],
    current_env: Optional[str],
    conf: "Config",
    loader: "IniLoader",
) -> Iterator[Union[SectionProxy, ConfigSet]]:
    # if we have an env name specified take only from there
    # config is None only when loading the global tox config file for the CLI arguments, in this case no replace works
    if env is not None:
        if env in conf:
            yield conf.get_env(env)
        else:
            raise KeyError(f"missing tox environment with name {env}")

    if section is None:
        # if no section specified perhaps it's an unregistered config:
        # 1. try first from core conf
        yield conf.core
        # 2. and then fallback to our own environment
        if current_env is not None:
            yield conf.get_env(current_env)
        return

    # if there's a section, special handle the core section under name tox
    if section == loader.core_prefix:
        yield conf.core  # try via registered configs
    value = loader.get_section(section)  # fallback to section
    if value is not None:
        yield value


def replace_pos_args(conf: "Config", env_name: Optional[str], args: List[str]) -> str:
    to_path: Optional[Path] = None
    if env_name is not None:  # pragma: no branch
        env_conf = conf.get_env(env_name)
        try:
            if env_conf["args_are_paths"]:  # pragma: no branch
                to_path = env_conf["change_dir"]
        except KeyError:
            pass
    pos_args = conf.pos_args(to_path)
    if pos_args is None:
        replace_value = ":".join(args)  # if we use the defaults join back remaining args
    else:
        replace_value = shell_cmd(pos_args)
    return replace_value


def replace_env(conf: "Config", env_name: Optional[str], args: List[str], chain: List[str]) -> str:
    key = args[0]
    new_key = f"env:{key}"

    if env_name is not None:  # on core no set env support # pragma: no branch
        if new_key not in chain:  # check if set env
            chain.append(new_key)
            env_conf = conf.get_env(env_name)
            set_env: SetEnv = env_conf["set_env"]
            if key in set_env:
                return set_env.load(key, chain)
        elif chain[-1] != new_key:  # if there's a chain but only self-refers than use os.environ
            raise ValueError(f"circular chain between set env {', '.join(i[4:] for i in chain[chain.index(new_key):])}")

    if key in os.environ:
        return os.environ[key]

    return "" if len(args) == 1 else args[1]


def replace_tty(args: List[str]) -> str:
    if sys.stdout.isatty():
        result = args[0] if len(args) > 0 else ""
    else:
        result = args[1] if len(args) > 1 else ""
    return result


__all__ = (
    "BASE_TEST_ENV",
    "replace",
    "find_replace_part",
)



from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

from .of_type import ConfigConstantDefinition, ConfigDefinition, ConfigDynamicDefinition
from .set_env import SetEnv
from .types import EnvList

if TYPE_CHECKING:
    from tox.config.loader.api import Loader
    from tox.config.main import Config

V = TypeVar("V")


class ConfigSet:
    """A set of configuration that belong together (such as a tox environment settings, core tox settings)"""

    def __init__(self, conf: "Config", name: Optional[str]):
        self._name = name
        self._conf = conf
        self.loaders: List[Loader[Any]] = []
        self._defined: Dict[str, ConfigDefinition[Any]] = {}
        self._keys: Dict[str, None] = {}
        self._alias: Dict[str, str] = {}

    def add_config(
        self,
        keys: Union[str, Sequence[str]],
        of_type: Type[V],
        default: Union[Callable[["Config", Optional[str]], V], V],
        desc: str,
        post_process: Optional[Callable[[V], V]] = None,
        kwargs: Optional[Mapping[str, Any]] = None,
    ) -> ConfigDynamicDefinition[V]:
        """
        Add configuration value.

        :param keys: the keys under what to register the config (first is primary key)
        :param of_type: the type of the config value
        :param default: the default value of the config value
        :param desc: a help message describing the configuration
        :param post_process: a callback to post-process the configuration value after it has been loaded
        :param kwargs: additional arguments to pass to the configuration type at construction time
        :return: the new dynamic config definition
        """
        keys_ = self._make_keys(keys)
        definition = ConfigDynamicDefinition(keys_, desc, self._name, of_type, default, post_process, kwargs)
        result = self._add_conf(keys_, definition)
        return cast(ConfigDynamicDefinition[V], result)

    def add_constant(self, keys: Union[str, Sequence[str]], desc: str, value: V) -> ConfigConstantDefinition[V]:
        """
        Add a constant value.

        :param keys: the keys under what to register the config (first is primary key)
        :param desc: a help message describing the configuration
        :param value: the config value to use
        :return: the new constant config value
        """
        keys_ = self._make_keys(keys)
        definition = ConfigConstantDefinition(keys_, desc, self._name, value)
        result = self._add_conf(keys_, definition)
        return cast(ConfigConstantDefinition[V], result)

    @staticmethod
    def _make_keys(keys: Union[str, Sequence[str]]) -> Sequence[str]:
        return (keys,) if isinstance(keys, str) else keys

    def _add_conf(self, keys: Sequence[str], definition: ConfigDefinition[V]) -> ConfigDefinition[V]:
        key = keys[0]
        if key in self._defined:
            earlier = self._defined[key]
            # core definitions may be defined multiple times as long as all their options match, first defined wins
            if self._name is None and definition == earlier:
                definition = earlier
            else:
                raise ValueError(f"config {key} already defined")
        else:
            self._keys[key] = None
            for item in keys:
                self._alias[item] = key
            for key in keys:
                self._defined[key] = definition
        return definition

    def __getitem__(self, item: str) -> Any:
        """
        Get the config value for a given key (will materialize in case of dynamic config).

        :param item: the config key
        :return: the configuration value
        """
        return self.load(item)

    def load(self, item: str, chain: Optional[List[str]] = None) -> Any:
        """
        Get the config value for a given key (will materialize in case of dynamic config).

        :param item: the config key
        :param chain: a chain of configuration keys already loaded for this load operation (used to detect circles)
        :return: the configuration value
        """
        config_definition = self._defined[item]
        if chain is None:
            chain = []
        env_name = "tox" if self._name is None else f"testenv:{self._name}"
        key = f"{env_name}.{item}"
        if key in chain:
            raise ValueError(f"circular chain detected {', '.join(chain[chain.index(key):])}")
        chain.append(key)
        return config_definition(self._conf, item, self.loaders, chain)

    def __repr__(self) -> str:
        values = (v for v in (f"name={self._name!r}" if self._name else "", f"loaders={self.loaders!r}") if v)
        return f"{self.__class__.__name__}({', '.join(values)})"

    def __iter__(self) -> Iterator[str]:
        """:return: iterate through the defined config keys (primary keys used)"""
        return iter(self._keys.keys())

    def __contains__(self, item: str) -> bool:
        """
        Check if a configuration key is within the config set.

        :param item: the configuration value
        :return: a boolean indicating the truthiness of the statement
        """
        return item in self._alias

    def unused(self) -> List[str]:
        """:return: Return a list of keys present in the config source but not used"""
        found: Set[str] = set()
        # keys within loaders (only if the loader is not a parent too)
        parents = {id(i.parent) for i in self.loaders if i.parent is not None}
        for loader in self.loaders:
            if id(loader) not in parents:
                found.update(loader.found_keys())
        found -= self._defined.keys()
        return sorted(found)

    def primary_key(self, key: str) -> str:
        """
        Get the primary key for a config key.

        :param key: the config key
        :return: the key that's considered the primary for the input key
        """
        return self._alias[key]


class CoreConfigSet(ConfigSet):
    """Configuration set for the core tox config"""

    def __init__(self, conf: "Config", root: Path) -> None:
        super().__init__(conf, name=None)
        self.add_config(
            keys=["tox_root", "toxinidir"],
            of_type=Path,
            default=root,
            desc="the root directory (where the configuration file is found)",
        )

        def work_dir_builder(conf: "Config", env_name: Optional[str]) -> Path:  # noqa
            # here we pin to .tox/4 to be able to use in parallel with v3 until final release
            return (conf.work_dir if conf.work_dir is not None else cast(Path, self["tox_root"])) / ".tox" / "4"

        self.add_config(
            keys=["work_dir", "toxworkdir"],
            of_type=Path,
            default=work_dir_builder,
            desc="working directory",
        )
        self.add_config(
            keys=["temp_dir"],
            of_type=Path,
            default=lambda conf, _: cast(Path, self["tox_root"]) / ".temp",
            desc="temporary directory cleaned at start",
        )
        self.add_config(
            keys=["env_list", "envlist"],
            of_type=EnvList,
            default=EnvList([]),
            desc="define environments to automatically run",
        )


class EnvConfigSet(ConfigSet):
    """Configuration set for a tox environment"""

    def __init__(self, conf: "Config", name: Optional[str]):
        super().__init__(conf, name=name)
        self.default_set_env_loader: Callable[[], Mapping[str, str]] = lambda: {}

        def set_env_post_process(values: SetEnv) -> SetEnv:
            values.update_if_not_present(self.default_set_env_loader())
            return values

        self.add_config(
            keys=["set_env", "setenv"],
            of_type=SetEnv,
            default=SetEnv(""),
            desc="environment variables to set when running commands in the tox environment",
            post_process=set_env_post_process,
        )

    @property
    def name(self) -> str:
        return self._name  # type: ignore


__all__ = (
    "ConfigSet",
    "CoreConfigSet",
    "EnvConfigSet",
)



import os
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Sequence, Tuple

from tox.config.loader.api import Loader, Override, OverrideMap

from .sets import CoreConfigSet, EnvConfigSet
from .source import Source

if TYPE_CHECKING:
    from .cli.parser import Parsed


class Config:
    """Main configuration object for tox."""

    def __init__(
        self,
        config_source: Source,
        overrides: List[Override],
        root: Path,
        pos_args: Optional[Sequence[str]],
        work_dir: Path,
    ) -> None:
        self._pos_args = None if pos_args is None else tuple(pos_args)
        self._work_dir = work_dir
        self._root = root

        self._overrides: OverrideMap = defaultdict(list)
        for override in overrides:
            self._overrides[override.namespace].append(override)

        self._src = config_source
        self._env_to_set: Dict[str, EnvConfigSet] = OrderedDict()
        self._core_set: Optional[CoreConfigSet] = None
        self.register_config_set: Callable[[str, EnvConfigSet], Any] = lambda n, e: None

    def pos_args(self, to_path: Optional[Path]) -> Optional[Tuple[str, ...]]:
        """
        :param to_path: if not None rewrite relative posargs paths from cwd to to_path
        :return: positional argument
        """
        if self._pos_args is not None and to_path is not None and Path.cwd() != to_path:
            args = []
            to_path_str = os.path.abspath(str(to_path))  # we use os.path to unroll .. in path without resolve
            for arg in self._pos_args:
                path_arg = Path(arg)
                if path_arg.exists() and not path_arg.is_absolute():
                    path_arg_str = os.path.abspath(str(path_arg))  # we use os.path to unroll .. in path without resolve
                    relative = os.path.relpath(path_arg_str, to_path_str)  # we use os.path to not fail when not within
                    args.append(relative)
                else:
                    args.append(arg)
            return tuple(args)
        return self._pos_args

    @property
    def work_dir(self) -> Path:
        """:return: working directory for this project"""
        return self._work_dir

    @property
    def src_path(self) -> Path:
        """:return: the location of the tox configuration source"""
        return self._src.path

    def __iter__(self) -> Iterator[str]:
        """:return: an iterator that goes through existing environments"""
        return self._src.envs(self.core)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(config_source={self._src!r})"

    def __contains__(self, item: str) -> bool:
        """:return: check if an environment already exists"""
        return any(name for name in self if name == item)

    @classmethod
    def make(cls, parsed: "Parsed", pos_args: Optional[Sequence[str]], source: Source) -> "Config":
        """Make a tox configuration object."""
        # root is the project root, where the configuration file is at
        # work dir is where we put our own files
        root: Path = source.path.parent if parsed.root_dir is None else parsed.root_dir
        work_dir: Path = source.path.parent if parsed.work_dir is None else parsed.work_dir
        return cls(
            config_source=source,
            overrides=parsed.override,
            pos_args=pos_args,
            root=root,
            work_dir=work_dir,
        )

    @property
    def core(self) -> CoreConfigSet:
        """:return: the core configuration"""
        if self._core_set is not None:
            return self._core_set
        core = CoreConfigSet(self, self._root)
        for loader in self._src.get_core(self._overrides):
            core.loaders.append(loader)

        from tox.plugin.manager import MANAGER

        MANAGER.tox_add_core_config(core)
        self._core_set = core
        return core

    def get_env(
        self, item: str, package: bool = False, loaders: Optional[Sequence[Loader[Any]]] = None
    ) -> EnvConfigSet:
        """
        Return the configuration for a given tox environment (will create if not exist yet).

        :param item: the name of the environment
        :param package: a flag indicating if the environment is of type packaging or not (only used for creation)
        :param loaders: loaders to use for this configuration (only used for creation)
        :return: the tox environments config
        """
        try:
            return self._env_to_set[item]
        except KeyError:
            env = EnvConfigSet(self, item)
            self._env_to_set[item] = env
            if loaders is not None:
                env.loaders.extend(loaders)
            for loader in self._src.get_env_loaders(item, self._overrides, package, env):
                env.loaders.append(loader)
            # whenever we load a new configuration we need to build a tox environment which process defines the valid
            # configuration values
            self.register_config_set(item, env)
            return env



"""
A tox environment that can build packages.
"""
from abc import ABC, abstractmethod
from pathlib import Path
from threading import Lock
from typing import Any, Generator, List, Optional, Set, Tuple, cast

from tox.config.main import Config
from tox.config.sets import EnvConfigSet

from .api import ToxEnv, ToxEnvCreateArgs


class Package:
    """package"""


class PathPackage(Package):
    def __init__(self, path: Path) -> None:
        super().__init__()
        self.path = path

    def __str__(self) -> str:
        return str(self.path)


class PackageToxEnv(ToxEnv, ABC):
    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        super().__init__(create_args)
        self._envs: Set[str] = set()
        self._lock = Lock()

    def register_config(self) -> None:
        super().register_config()
        self.core.add_config(
            keys=["package_root", "setupdir"],
            of_type=Path,
            default=cast(Path, self.core["tox_root"]),
            desc="indicates where the packaging root file exists (historically setup.py file or pyproject.toml now)",
        )
        self.conf.add_config(
            keys=["package_root", "setupdir"],
            of_type=Path,
            default=cast(Path, self.core["package_root"]),
            desc="indicates where the packaging root file exists (historically setup.py file or pyproject.toml now)",
        )

    def _recreate_default(self, conf: "Config", value: Optional[str]) -> bool:
        return self.options.no_recreate_pkg is False and super()._recreate_default(conf, value)

    def create_package_env(
        self, name: str, info: Tuple[Any, ...]  # noqa: U100
    ) -> Generator[Tuple[str, str], "PackageToxEnv", None]:
        """allow creating sub-package envs"""

    @abstractmethod
    def perform_packaging(self, for_env: EnvConfigSet) -> List[Package]:  # noqa: U100
        raise NotImplementedError

    def notify_of_run_env(self, conf: EnvConfigSet) -> None:
        with self._lock:
            self._envs.add(conf.name)

    def teardown_env(self, conf: EnvConfigSet) -> None:
        with self._lock:
            self._envs.remove(conf.name)
            has_envs = bool(self._envs)
        if not has_envs:
            self._teardown()



import sys
from typing import Callable, Tuple

import pytest

from tox.pytest import ToxProjectCreator


@pytest.mark.parametrize("has_prev", [True, False])
def test_depends(
    tox_project: ToxProjectCreator, patch_prev_py: Callable[[bool], Tuple[str, str]], has_prev: bool
) -> None:
    prev_ver, impl = patch_prev_py(has_prev)
    ver = sys.version_info[0:2]
    py = f"py{''.join(str(i) for i in ver)}"
    prev_py = f"py{prev_ver}"
    project = tox_project(
        {
            "tox.ini": f"""
    [tox]
    env_list = py,{py},{prev_py},py31,cov2,cov
    [testenv]
    package = wheel
    [testenv:cov]
    depends = py,{py},{prev_py},py31
    skip_install = true
    [testenv:cov2]
    depends = cov
    skip_install = true
    """
        }
    )
    outcome = project.run("de")
    outcome.assert_success()
    lines = outcome.out.splitlines()
    assert lines[0] == f"Execution order: py, {py},{f' {prev_py},' if has_prev else '' } cov, cov2"
    expected_lines = [
        "ALL",
        "   py ~ .pkg",
        f"   {py} ~ .pkg",
    ]
    if has_prev:
        expected_lines.append(f"   {prev_py} ~ .pkg | .pkg-{impl}{prev_ver}")
    expected_lines.extend(
        [
            "   cov2",
            "      cov",
            "         py ~ .pkg",
            f"         {py} ~ .pkg",
        ]
    )
    if has_prev:
        expected_lines.append(f"         {prev_py} ~ .pkg | .pkg-{impl}{prev_ver}")
    expected_lines.extend(
        [
            "   cov",
            "      py ~ .pkg",
            f"      {py} ~ .pkg",
        ]
    )
    if has_prev:
        expected_lines.append(f"      {prev_py} ~ .pkg | .pkg-{impl}{prev_ver}")
    assert lines[1:] == expected_lines



"""
A tox build environment that handles Python packages.
"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Sequence, Tuple

from packaging.requirements import Requirement

from ..package import Package, PackageToxEnv, PathPackage
from .api import Python


class PythonPackage(Package):
    """python package"""


class PythonPathPackageWithDeps(PathPackage):
    def __init__(self, path: Path, deps: Sequence[Any]) -> None:
        super().__init__(path=path)
        self.deps: Sequence[Package] = deps


class WheelPackage(PythonPathPackageWithDeps):
    """wheel package"""


class SdistPackage(PythonPathPackageWithDeps):
    """sdist package"""


class DevLegacyPackage(PythonPathPackageWithDeps):
    """legacy dev package"""


class PythonPackageToxEnv(Python, PackageToxEnv, ABC):
    def register_config(self) -> None:
        super().register_config()

    def _setup_env(self) -> None:
        """setup the tox environment"""
        super()._setup_env()
        self.installer.install(self.requires(), PythonPackageToxEnv.__name__, "requires")

    @abstractmethod
    def requires(self) -> Tuple[Requirement, ...]:
        raise NotImplementedError



from argparse import ArgumentParser
from typing import Any, Iterator, List, Optional, Union

from tox.config.loader.str_convert import StrConvert


class CliEnv:
    def __init__(self, value: Union[None, List[str], str] = None):
        if isinstance(value, str):
            value = StrConvert().to(value, of_type=List[str], kwargs={})
        self._names = value

    def __iter__(self) -> Iterator[str]:
        if self._names is not None:  # pragma: no branch
            yield from self._names

    def __str__(self) -> str:
        if self.all:
            return "ALL"
        if self.use_default_list:
            return "<env_list>"
        return ",".join(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({'' if self.use_default_list else repr(str(self))})"

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self._names == other._names

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    @property
    def all(self) -> bool:
        return "ALL" in self

    @property
    def use_default_list(self) -> bool:
        return len(list(self)) == 0


def env_list_flag(parser: ArgumentParser, default: Optional[CliEnv] = None, multiple: bool = True) -> None:
    help_msg = (
        "tox environment(s) to run (ALL -> all environments, not set -> <env_list>)"
        if multiple
        else "tox environment to run"
    )
    parser.add_argument("-e", dest="env", help=help_msg, default=CliEnv() if default is None else default, type=CliEnv)



"""
A tox run environment that handles the Python language.
"""
from abc import ABC
from pathlib import Path
from typing import Iterator, List, Optional, Set, Tuple

from tox.config.main import Config
from tox.report import HandledError
from tox.tox_env.errors import Skip
from tox.tox_env.package import Package, PathPackage
from tox.tox_env.python.package import PythonPackageToxEnv
from tox.tox_env.python.pip.req_file import PythonDeps

from ..api import ToxEnvCreateArgs
from ..runner import RunToxEnv
from .api import Python


class PythonRun(Python, RunToxEnv, ABC):
    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        super().__init__(create_args)

    def register_config(self) -> None:
        super().register_config()
        deps_kwargs = {"root": self.core["toxinidir"]}
        self.conf.add_config(
            keys="deps",
            of_type=PythonDeps,
            kwargs=deps_kwargs,
            default=PythonDeps("", **deps_kwargs),
            desc="Name of the python dependencies as specified by PEP-440",
        )
        self.core.add_config(
            keys=["skip_missing_interpreters"],
            default=True,
            of_type=bool,
            desc="skip running missing interpreters",
        )

    def iter_package_env_types(self) -> Iterator[Tuple[str, str, str]]:
        yield from super().iter_package_env_types()
        if self.pkg_type == "wheel":
            wheel_build_env: str = self.conf["wheel_build_env"]
            if wheel_build_env not in self._package_envs:  # pragma: no branch
                package_tox_env_type = self.conf["package_tox_env_type"]
                yield "wheel", wheel_build_env, package_tox_env_type

    @property
    def _package_types(self) -> Tuple[str, ...]:
        return "wheel", "sdist", "dev-legacy", "skip"

    def _register_package_conf(self) -> bool:
        desc = f"package installation mode - {' | '.join(i for i in self._package_types)} "
        if not super()._register_package_conf():
            self.conf.add_constant(["package"], desc, "skip")
            return False
        self.conf.add_config(keys=["use_develop", "usedevelop"], desc="use develop mode", default=False, of_type=bool)
        develop_mode = self.conf["use_develop"] or getattr(self.options, "develop", False)
        if develop_mode:
            self.conf.add_constant(["package"], desc, "dev-legacy")
        else:
            self.conf.add_config(keys="package", of_type=str, default="sdist", desc=desc)
        pkg_type = self.pkg_type

        if pkg_type == "skip":
            return False

        if pkg_type == "wheel":

            def default_wheel_tag(conf: "Config", env_name: Optional[str]) -> str:
                # https://www.python.org/dev/peps/pep-0427/#file-name-convention
                # when building wheels we need to ensure that the built package is compatible with the target env
                # compatibility is documented within https://www.python.org/dev/peps/pep-0427/#file-name-convention
                # a wheel tag example: {distribution}-{version}(-{build tag})?-{python tag}-{abi tag}-{platform tag}.whl
                # python only code are often compatible at major level (unless universal wheel in which case both 2/3)
                # c-extension codes are trickier, but as of today both poetry/setuptools uses pypa/wheels logic
                # https://github.com/pypa/wheel/blob/master/src/wheel/bdist_wheel.py#L234-L280
                default_package_env = self._package_envs["default"]
                self_py = self.base_python
                if self_py is not None and isinstance(default_package_env, PythonPackageToxEnv):
                    default_pkg_py = default_package_env.base_python
                    if (
                        default_pkg_py.version_no_dot == self_py.version_no_dot
                        and default_pkg_py.impl_lower == self_py.impl_lower
                    ):
                        return default_package_env.conf.name
                if self_py is None:
                    raise ValueError(f"could not resolve base python for {self.conf.name}")
                return f"{default_package_env.conf.name}-{self_py.impl_lower}{self_py.version_no_dot}"

            self.conf.add_config(
                keys=["wheel_build_env"],
                of_type=str,
                default=default_wheel_tag,
                desc="wheel tag to use for building applications",
            )
        self.conf.add_config(
            keys=["extras"],
            of_type=Set[str],
            default=set(),
            desc="extras to install of the target package",
        )
        return True

    @property
    def pkg_type(self) -> str:
        pkg_type: str = self.conf["package"]
        if pkg_type not in self._package_types:
            values = ", ".join(self._package_types)
            raise HandledError(f"invalid package config type {pkg_type} requested, must be one of {values}")
        return pkg_type

    def _setup_env(self) -> None:
        super()._setup_env()
        self._install_deps()

    def _install_deps(self) -> None:
        requirements_file: PythonDeps = self.conf["deps"]
        self.installer.install(requirements_file, PythonRun.__name__, "deps")

    def _build_packages(self) -> List[Package]:
        explicit_install_package: Optional[Path] = getattr(self.options, "install_pkg", None)
        if explicit_install_package is not None:
            return [PathPackage(explicit_install_package)]

        package_env = self._package_envs[self._get_package_env()]
        with package_env.display_context(self._has_display_suspended):
            try:
                packages = package_env.perform_packaging(self.conf)
            except Skip as exception:
                raise Skip(f"{exception.args[0]} for package environment {package_env.conf['env_name']}")
        return packages

    def _get_package_env(self) -> str:
        return "wheel" if self.pkg_type == "wheel" else "default"



import json
import os
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from subprocess import check_call
from typing import Iterator, List, Optional
from zipfile import ZipFile

import pytest
from filelock import FileLock
from packaging.requirements import Requirement

from tox.pytest import Index, IndexServer, MonkeyPatch, TempPathFactory, ToxProjectCreator

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution  # noqa

ROOT = Path(__file__).parents[1]


@contextmanager
def elapsed(msg: str) -> Iterator[None]:
    start = time.monotonic()
    try:
        yield
    finally:
        print(f"done in {time.monotonic() - start}s {msg}")


@pytest.fixture(scope="session")
def tox_wheel(tmp_path_factory: TempPathFactory, worker_id: str) -> Path:
    if worker_id == "master":  # if not running under xdist we can just return
        return _make_tox_wheel(tmp_path_factory)  # pragma: no cover
    # otherwise we need to ensure only one worker creates the wheel, and the rest reuses
    root_tmp_dir = tmp_path_factory.getbasetemp().parent
    cache_file = root_tmp_dir / "tox_wheel.json"
    with FileLock(f"{cache_file}.lock"):
        if cache_file.is_file():
            data = Path(json.loads(cache_file.read_text()))  # pragma: no cover
        else:
            data = _make_tox_wheel(tmp_path_factory)
            cache_file.write_text(json.dumps(str(data)))
    return data


def _make_tox_wheel(tmp_path_factory: TempPathFactory) -> Path:
    with elapsed("acquire current tox wheel"):  # takes around 3.2s on build
        package: Optional[Path] = None
        if "TOX_PACKAGE" in os.environ:
            env_tox_pkg = Path(os.environ["TOX_PACKAGE"])  # pragma: no cover
            if env_tox_pkg.exists() and env_tox_pkg.suffix == ".whl":  # pragma: no cover
                package = env_tox_pkg  # pragma: no cover
        if package is None:
            # when we don't get a wheel path injected, build it (for example when running from an IDE)
            into = tmp_path_factory.mktemp("dist")  # pragma: no cover
            package = build_wheel(into, Path(__file__).parents[1], isolation=False)  # pragma: no cover
        return package


@pytest.fixture(scope="session")
def tox_wheels(tox_wheel: Path, tmp_path_factory: TempPathFactory) -> List[Path]:
    with elapsed("acquire dependencies for current tox"):  # takes around 1.5s if already cached
        result: List[Path] = [tox_wheel]
        info = tmp_path_factory.mktemp("info")
        with ZipFile(str(tox_wheel), "r") as zip_file:
            zip_file.extractall(path=info)
        dist_info = next((i for i in info.iterdir() if i.suffix == ".dist-info"), None)
        if dist_info is None:  # pragma: no cover
            raise RuntimeError(f"no tox.dist-info inside {tox_wheel}")
        distribution = Distribution.at(dist_info)
        wheel_cache = ROOT / ".wheel_cache" / f"{sys.version_info.major}.{sys.version_info.minor}"
        wheel_cache.mkdir(parents=True, exist_ok=True)
        cmd = [sys.executable, "-I", "-m", "pip", "download", "-d", str(wheel_cache)]
        for req in distribution.requires:
            requirement = Requirement(req)
            if not requirement.extras:  # pragma: no branch  # we don't need to install any extras (tests/docs/etc)
                cmd.append(req)
        check_call(cmd)
        result.extend(wheel_cache.iterdir())
        return result


@pytest.fixture(scope="session")
def demo_pkg_inline_wheel(tmp_path_factory: TempPathFactory, demo_pkg_inline: Path) -> Path:
    return build_wheel(tmp_path_factory.mktemp("dist"), demo_pkg_inline)


def build_wheel(dist_dir: Path, of: Path, isolation: bool = True) -> Path:
    from build.__main__ import build_package  # noqa

    build_package(str(of), str(dist_dir), distributions=["wheel"], isolation=isolation)
    package = next(dist_dir.iterdir())
    return package


@pytest.fixture(scope="session")
def pypi_index_self(pypi_server: IndexServer, tox_wheels: List[Path], demo_pkg_inline_wheel: Path) -> Index:
    with elapsed("start devpi and create index"):  # takes around 1s
        self_index = pypi_server.create_index("self", "volatile=False")
    with elapsed("upload tox and its wheels to devpi"):  # takes around 3.2s on build
        self_index.upload(tox_wheels + [demo_pkg_inline_wheel])
    return self_index


def test_provision_requires_nok(tox_project: ToxProjectCreator) -> None:
    ini = "[tox]\nrequires = pkg-does-not-exist\n setuptools==1\nskipsdist=true\n"
    outcome = tox_project({"tox.ini": ini}).run("c", "-e", "py")
    outcome.assert_failed()
    outcome.assert_out_err(
        r".*will run in automatically provisioned tox, host .* is missing \[requires \(has\)\]:"
        r" pkg-does-not-exist \(N/A\), setuptools==1 \(.*\).*",
        r".*",
        regex=True,
    )


@pytest.mark.integration()
def test_provision_requires_ok(
    tox_project: ToxProjectCreator, pypi_index_self: Index, monkeypatch: MonkeyPatch, tmp_path: Path
) -> None:
    pypi_index_self.use(monkeypatch)
    proj = tox_project({"tox.ini": "[tox]\nrequires=demo-pkg-inline\n[testenv]\npackage=skip"})
    log = tmp_path / "out.log"

    # initial run
    result_first = proj.run("r", "--result-json", str(log))
    result_first.assert_success()
    prov_msg = (
        f"ROOT: will run in automatically provisioned tox, host {sys.executable} is missing"
        f" [requires (has)]: demo-pkg-inline (N/A)"
    )
    assert prov_msg in result_first.out

    with log.open("rt") as file_handler:
        log_report = json.load(file_handler)
    assert "py" in log_report["testenvs"]

    # recreate without recreating the provisioned env
    provision_env = result_first.env_conf(".tox")["env_dir"]
    result_recreate_no_pr = proj.run("r", "--recreate", "--no-recreate-provision")
    result_recreate_no_pr.assert_success()
    assert prov_msg in result_recreate_no_pr.out
    assert f"ROOT: remove tox env folder {provision_env}" not in result_recreate_no_pr.out, result_recreate_no_pr.out

    # recreate with recreating the provisioned env
    result_recreate = proj.run("r", "--recreate")
    result_recreate.assert_success()
    assert prov_msg in result_recreate.out
    assert f"ROOT: remove tox env folder {provision_env}" in result_recreate.out, result_recreate.out


@pytest.mark.integration()
def test_provision_platform_check(
    tox_project: ToxProjectCreator, pypi_index_self: Index, monkeypatch: MonkeyPatch
) -> None:
    pypi_index_self.use(monkeypatch)
    ini = "[tox]\nrequires=demo-pkg-inline\n[testenv]\npackage=skip\n[testenv:.tox]\nplatform=wrong_platform"
    proj = tox_project({"tox.ini": ini})

    result = proj.run("r")
    result.assert_failed(-2)
    msg = f"cannot provision tox environment .tox because platform {sys.platform} does not match wrong_platform"
    assert msg in result.out



from pathlib import Path

from .ini import IniSource


class SetupCfg(IniSource):
    """Configuration sourced from a tox.ini file"""

    CORE_PREFIX = "tox:tox"
    FILENAME = "setup.cfg"

    def __init__(self, path: Path):
        super().__init__(path)
        if not self._parser.has_section(self.CORE_PREFIX):
            raise ValueError


__all__ = ("SetupCfg",)



from collections import OrderedDict
from typing import Any, Iterator, List, Sequence

from tox.execute.request import shell_cmd


class Command:
    """A command to execute."""

    def __init__(self, args: List[str]) -> None:
        """
        Create a new command to execute

        :param args: the command line arguments (first value can be ``-`` to indicate ignore the exit code)
        """
        self.ignore_exit_code: bool = args[0] == "-"  #: a flag indicating if the exit code should be ignored
        self.args: List[str] = args[1:] if self.ignore_exit_code else args  #: the command line arguments

    def __repr__(self) -> str:
        return f"{type(self).__name__}(args={(['-'] if self.ignore_exit_code else [])+ self.args!r})"

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and (self.args, self.ignore_exit_code) == (other.args, other.ignore_exit_code)

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    @property
    def shell(self) -> str:
        """:return: a shell representation of the command (platform dependent)"""
        return shell_cmd(self.args)


class EnvList:
    """A tox environment list"""

    def __init__(self, envs: Sequence[str]) -> None:
        """
        Crate a new tox environment list.

        :param envs: the list of tox environments
        """
        self.envs = list(OrderedDict((e, None) for e in envs).keys())

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.envs!r})"

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.envs == other.envs

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __iter__(self) -> Iterator[str]:
        """:return: iterator that goes through the defined env-list"""
        return iter(self.envs)

    def __bool__(self) -> bool:
        """:return: ``True`` if there are any environments defined"""
        return bool(self.envs)


__all__ = (
    "Command",
    "EnvList",
)



from typing import Callable, Dict, Iterator, List, Mapping, Optional, Tuple

Replacer = Callable[[str, List[str]], str]


class SetEnv:
    def __init__(self, raw: str) -> None:
        self.replacer: Replacer = lambda s, c: s
        self._later: List[str] = []
        self._raw: Dict[str, str] = {}
        from .loader.ini.replace import find_replace_part

        for line in raw.splitlines():
            if line.strip():
                try:
                    key, value = self._extract_key_value(line)
                    if "{" in key:
                        raise ValueError(f"invalid line {line!r} in set_env")
                except ValueError:
                    _, __, match = find_replace_part(line, 0)
                    if match:
                        self._later.append(line)
                    else:
                        raise
                else:
                    self._raw[key] = value
        self._materialized: Dict[str, str] = {}

    @staticmethod
    def _extract_key_value(line: str) -> Tuple[str, str]:
        key, sep, value = line.partition("=")
        if sep:
            return key.strip(), value.strip()
        else:
            raise ValueError(f"invalid line {line!r} in set_env")

    def load(self, item: str, chain: Optional[List[str]] = None) -> str:
        if chain is None:
            chain = [f"env:{item}"]
        if item in self._materialized:
            return self._materialized[item]
        raw = self._raw[item]
        result = self.replacer(raw, chain)  # apply any replace options
        result = result.replace(r"\#", "#")  # unroll escaped comment with replacement
        self._materialized[item] = result
        self._raw.pop(item, None)  # if the replace requires the env we may be called again, so allow pop to fail
        return result

    def __contains__(self, item: object) -> bool:
        return isinstance(item, str) and item in self.__iter__()

    def __iter__(self) -> Iterator[str]:
        # start with the materialized ones, maybe we don't need to materialize the raw ones
        yield from self._materialized.keys()
        yield from list(self._raw.keys())  # iterating over this may trigger materialization and change the dict
        while self._later:
            line = self._later.pop(0)
            expanded_line = self.replacer(line, [])
            sub_raw = dict(self._extract_key_value(sub_line) for sub_line in expanded_line.splitlines() if sub_line)
            self._raw.update(sub_raw)
            yield from sub_raw.keys()

    def update_if_not_present(self, param: Mapping[str, str]) -> None:
        for key, value in param.items():
            # do not override something already set explicitly
            if key not in self._raw and key not in self._materialized:
                self._materialized[key] = value


__all__ = ("SetEnv",)



"""Sources."""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Iterator

from tox.config.loader.api import Loader, OverrideMap

from ..sets import ConfigSet, CoreConfigSet


class Source(ABC):
    """
    Source is able to return a configuration value (for either the core or per environment source).
    """

    FILENAME = ""

    def __init__(self, path: Path) -> None:
        self.path: Path = path  #: the path to the configuration source

    @abstractmethod
    def get_core(self, override_map: OverrideMap) -> Iterator[Loader[Any]]:  # noqa: U100
        """
        Return a loader that loads the core configuration values.

        :param override_map: a list of overrides to apply
        :returns: the core loader from this source
        """
        raise NotImplementedError

    @abstractmethod
    def get_env_loaders(
        self, env_name: str, override_map: OverrideMap, package: bool, conf: ConfigSet  # noqa: U100
    ) -> Iterator[Loader[Any]]:
        """
        Return the load for this environment.

        :param env_name: the environment name
        :param override_map: a list of overrides to apply
        :param package: a flag indicating if this is a package environment, otherwise is of type run
        :param conf: the config set to use
        :returns: an iterable of loaders extracting config value from this source
        """
        raise NotImplementedError

    @abstractmethod
    def envs(self, core_conf: "CoreConfigSet") -> Iterator[str]:  # noqa: U100
        """
        :param core_conf: the core configuration set
        :returns: a list of environments defined within this source
        """
        raise NotImplementedError


__all__ = ("Source",)



"""
Declare the abstract base class for tox environments that handle the Python language via the virtualenv project.
"""
import sys
from abc import ABC
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from virtualenv import __version__ as virtualenv_version
from virtualenv import session_via_cli
from virtualenv.create.creator import Creator
from virtualenv.run.session import Session

from tox.config.loader.str_convert import StrConvert
from tox.execute.api import Execute
from tox.execute.local_sub_process import LocalSubProcessExecutor
from tox.tox_env.python.pip.pip_install import Pip

from ...api import ToxEnvCreateArgs
from ..api import Python, PythonInfo


class VirtualEnv(Python, ABC):
    """A python executor that uses the virtualenv project with pip"""

    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        self._virtualenv_session: Optional[Session] = None
        self._executor: Optional[Execute] = None
        self._installer: Optional[Pip] = None
        super().__init__(create_args)

    def register_config(self) -> None:
        super().register_config()
        self.conf.add_config(
            keys=["system_site_packages", "sitepackages"],
            of_type=bool,
            default=lambda conf, name: StrConvert().to_bool(
                self._environment_variables.get("VIRTUALENV_SYSTEM_SITE_PACKAGES", "False")
            ),
            desc="create virtual environments that also have access to globally installed packages.",
        )
        self.conf.add_config(
            keys=["always_copy", "alwayscopy"],
            of_type=bool,
            default=lambda conf, name: StrConvert().to_bool(
                self._environment_variables.get(
                    "VIRTUALENV_COPIES", self._environment_variables.get("VIRTUALENV_ALWAYS_COPY", "False")
                )
            ),
            desc="force virtualenv to always copy rather than symlink",
        )
        self.conf.add_config(
            keys=["download"],
            of_type=bool,
            default=lambda conf, name: StrConvert().to_bool(
                self._environment_variables.get("VIRTUALENV_DOWNLOAD", "False")
            ),
            desc="true if you want virtualenv to upgrade pip/wheel/setuptools to the latest version",
        )

    @property
    def executor(self) -> Execute:
        if self._executor is None:
            self._executor = LocalSubProcessExecutor(self.options.is_colored)
        return self._executor

    @property
    def installer(self) -> Pip:
        if self._installer is None:
            self._installer = Pip(self)
        return self._installer

    def python_cache(self) -> Dict[str, Any]:
        base = super().python_cache()
        base.update(
            {
                "executable": str(self.base_python.extra["executable"]),
                "virtualenv version": virtualenv_version,
            }
        )
        return base

    def _get_env_journal_python(self) -> Dict[str, Any]:
        base = super()._get_env_journal_python()
        base["executable"] = str(self.base_python.extra["executable"])
        return base

    def _default_pass_env(self) -> List[str]:
        env = super()._default_pass_env()
        env.append("PIP_*")  # we use pip as installer
        env.append("VIRTUALENV_*")  # we use virtualenv as isolation creator
        return env

    def _default_set_env(self) -> Dict[str, str]:
        env = super()._default_set_env()
        env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
        return env

    @property
    def session(self) -> Session:
        if self._virtualenv_session is None:
            env_dir = [str(self.env_dir)]
            env = self.virtualenv_env_vars()
            self._virtualenv_session = session_via_cli(env_dir, options=None, setup_logging=False, env=env)
        return self._virtualenv_session

    def virtualenv_env_vars(self) -> Dict[str, str]:
        env = self._environment_variables.copy()
        base_python: List[str] = self.conf["base_python"]
        if "VIRTUALENV_NO_PERIODIC_UPDATE" not in env:
            env["VIRTUALENV_NO_PERIODIC_UPDATE"] = "True"
        site = getattr(self.options, "site_packages", False) or self.conf["system_site_packages"]
        env["VIRTUALENV_CLEAR"] = "False"
        env["VIRTUALENV_SYSTEM_SITE_PACKAGES"] = str(site)
        env["VIRTUALENV_COPIES"] = str(getattr(self.options, "always_copy", False) or self.conf["always_copy"])
        env["VIRTUALENV_DOWNLOAD"] = str(self.conf["download"])
        env["VIRTUALENV_PYTHON"] = "\n".join(base_python)
        return env

    @property
    def creator(self) -> Creator:
        return self.session.creator

    def create_python_env(self) -> None:
        self.session.run()

    def _get_python(self, base_python: List[str]) -> Optional[PythonInfo]:  # noqa: U100
        # the base pythons are injected into the virtualenv_env_vars, so we don't need to use it here
        try:
            interpreter = self.creator.interpreter
        except RuntimeError:  # if can't find
            return None
        return PythonInfo(
            implementation=interpreter.implementation,
            version_info=interpreter.version_info,
            version=interpreter.version,
            is_64=(interpreter.architecture == 64),
            platform=interpreter.platform,
            extra={"executable": Path(interpreter.system_executable)},
        )

    def prepend_env_var_path(self) -> List[Path]:
        """Paths to add to the executable"""
        # we use the original executable as shims may be somewhere else
        return list(dict.fromkeys((self.creator.bin_dir, self.creator.script_dir)))

    def env_site_package_dir(self) -> Path:
        return cast(Path, self.creator.purelib)

    def env_python(self) -> Path:
        return cast(Path, self.creator.exe)

    def env_bin_dir(self) -> Path:
        return cast(Path, self.creator.script_dir)

    @property
    def runs_on_platform(self) -> str:
        return sys.platform



import os
import sys
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path
from threading import RLock
from typing import Any, Dict, Iterator, List, NoReturn, Optional, Sequence, Set, Tuple, Union, cast

from cachetools import cached
from packaging.markers import Variable
from packaging.requirements import Requirement

from tox.config.sets import EnvConfigSet
from tox.execute.api import ExecuteStatus
from tox.execute.pep517_backend import LocalSubProcessPep517Executor
from tox.execute.request import StdinSource
from tox.plugin import impl
from tox.tox_env.api import ToxEnvCreateArgs
from tox.tox_env.errors import Fail
from tox.tox_env.package import Package
from tox.tox_env.python.package import DevLegacyPackage, PythonPackageToxEnv, SdistPackage, WheelPackage
from tox.tox_env.register import ToxEnvRegister
from tox.util.pep517.frontend import BackendFailed, CmdStatus, ConfigSettings, Frontend

from ..api import VirtualEnv

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution, PathDistribution  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution, PathDistribution  # noqa


TOX_PACKAGE_ENV_ID = "virtualenv-pep-517"


class ToxBackendFailed(Fail, BackendFailed):
    def __init__(self, backend_failed: BackendFailed) -> None:
        Fail.__init__(self)
        result: Dict[str, Any] = {
            "code": backend_failed.code,
            "exc_type": backend_failed.exc_type,
            "exc_msg": backend_failed.exc_msg,
        }
        BackendFailed.__init__(
            self,
            result,
            backend_failed.out,
            backend_failed.err,
        )


class ToxCmdStatus(CmdStatus):
    def __init__(self, execute_status: ExecuteStatus) -> None:
        self._execute_status = execute_status

    @property
    def done(self) -> bool:
        # 1. process died
        status = self._execute_status
        if status.exit_code is not None:  # pragma: no branch
            return True  # pragma: no cover
        # 2. the backend output reported back that our command is done
        return b"\n" in status.out.rpartition(b"Backend: Wrote response ")[0]

    def out_err(self) -> Tuple[str, str]:
        status = self._execute_status
        if status is None or status.outcome is None:  # interrupt before status create # pragma: no branch
            return "", ""  # pragma: no cover
        return status.outcome.out_err()


class Pep517VirtualEnvPackage(PythonPackageToxEnv, VirtualEnv, Frontend):
    """local file system python virtual environment via the virtualenv package"""

    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        VirtualEnv.__init__(self, create_args)
        root: Path = self.conf["package_root"]
        Frontend.__init__(self, *Frontend.create_args_from_folder(root))

        self._backend_executor_: Optional[LocalSubProcessPep517Executor] = None
        self._builds: Set[str] = set()
        self._distribution_meta: Optional[PathDistribution] = None
        self._package_dependencies: Optional[List[Requirement]] = None
        self._pkg_lock = RLock()  # can build only one package at a time
        into: Dict[str, Any] = {}
        pkg_cache = cached(into, key=lambda *args, **kwargs: "wheel" if "wheel_directory" in kwargs else "sdist")
        self.build_wheel = pkg_cache(self.build_wheel)  # type: ignore
        self.build_sdist = pkg_cache(self.build_sdist)  # type: ignore

    @staticmethod
    def id() -> str:
        return "virtualenv-pep-517"

    def register_config(self) -> None:
        super().register_config()
        self.conf.add_config(
            keys=["meta_dir"],
            of_type=Path,
            default=lambda conf, name: self.env_dir / ".meta",
            desc="directory where to put the project metadata files",
        )
        self.conf.add_config(
            keys=["pkg_dir"],
            of_type=Path,
            default=lambda conf, name: self.env_dir / "dist",
            desc="directory where to put project packages",
        )

    @property
    def pkg_dir(self) -> Path:
        return cast(Path, self.conf["pkg_dir"])

    @property
    def meta_folder(self) -> Path:
        meta_folder: Path = self.conf["meta_dir"]
        meta_folder.mkdir(exist_ok=True)
        return meta_folder

    def notify_of_run_env(self, conf: EnvConfigSet) -> None:
        super().notify_of_run_env(conf)
        self._builds.add(conf["package"])

    def _setup_env(self) -> None:
        super()._setup_env()
        if "wheel" in self._builds:
            build_requires = self.get_requires_for_build_wheel().requires
            self.installer.install(build_requires, PythonPackageToxEnv.__name__, "requires_for_build_wheel")
        if "sdist" in self._builds:
            build_requires = self.get_requires_for_build_sdist().requires
            self.installer.install(build_requires, PythonPackageToxEnv.__name__, "requires_for_build_sdist")

    def _teardown(self) -> None:
        if self._backend_executor_ is not None:
            try:
                if self._backend_executor.is_alive:
                    self._send("_exit")  # try first on amicable shutdown
            except SystemExit:  # if already has been interrupted ignore
                pass
            finally:
                self._backend_executor_.close()
        super()._teardown()

    def perform_packaging(self, for_env: EnvConfigSet) -> List[Package]:
        """build the package to install"""
        of_type: str = for_env["package"]
        extras: Set[str] = for_env["extras"]
        deps = self._dependencies_with_extras(self._get_package_dependencies(), extras)
        if of_type == "dev-legacy":
            deps = [*self.requires(), *self.get_requires_for_build_sdist().requires] + deps
            package: Package = DevLegacyPackage(self.core["tox_root"], deps)  # the folder itself is the package
        elif of_type == "sdist":
            with self._pkg_lock:
                package = SdistPackage(self.build_sdist(sdist_directory=self.pkg_dir).sdist, deps)
        elif of_type == "wheel":
            with self._pkg_lock:
                path = self.build_wheel(
                    wheel_directory=self.pkg_dir,
                    metadata_directory=self.meta_folder,
                    config_settings=self._wheel_config_settings,
                ).wheel
            package = WheelPackage(path, deps)
        else:  # pragma: no cover # for when we introduce new packaging types and don't implement
            raise TypeError(f"cannot handle package type {of_type}")  # pragma: no cover
        return [package]

    def _get_package_dependencies(self) -> List[Requirement]:
        with self._pkg_lock:
            if self._package_dependencies is None:  # pragma: no branch
                self._ensure_meta_present()
                requires: List[str] = cast(PathDistribution, self._distribution_meta).requires or []
                self._package_dependencies = [Requirement(i) for i in requires]
        return self._package_dependencies

    def _ensure_meta_present(self) -> None:
        if self._distribution_meta is not None:  # pragma: no branch
            return  # pragma: no cover
        self.setup()
        dist_info = self.prepare_metadata_for_build_wheel(self.meta_folder, self._wheel_config_settings).metadata
        self._distribution_meta = Distribution.at(str(dist_info))  # type: ignore[no-untyped-call]

    @staticmethod
    def _dependencies_with_extras(deps: List[Requirement], extras: Set[str]) -> List[Requirement]:
        result: List[Requirement] = []
        for req in deps:
            req = deepcopy(req)
            markers: List[Union[str, Tuple[Variable, Variable, Variable]]] = getattr(req.marker, "_markers", []) or []
            # find the extra marker (if has)
            _at: Optional[int] = None
            extra: Optional[str] = None
            for _at, (marker_key, op, marker_value) in (
                (_at_marker, marker)
                for _at_marker, marker in enumerate(markers)
                if isinstance(marker, tuple) and len(marker) == 3
            ):
                if marker_key.value == "extra" and op.value == "==":  # pragma: no branch
                    extra = marker_value.value
                    del markers[_at]
                    _at -= 1
                    if _at > 0 and (isinstance(markers[_at], str) and markers[_at] in ("and", "or")):
                        del markers[_at]
                    if len(markers) == 0:
                        req.marker = None
                    break
            if not (extra is None or extra in extras):
                continue
            result.append(req)
        return result

    @contextmanager
    def _wheel_directory(self) -> Iterator[Path]:
        yield self.pkg_dir  # use our local wheel directory for building wheel

    @property
    def _wheel_config_settings(self) -> Optional[ConfigSettings]:
        return {"--global-option": ["--bdist-dir", str(self.env_dir / "build")]}

    @property
    def _backend_executor(self) -> LocalSubProcessPep517Executor:
        if self._backend_executor_ is None:
            self._backend_executor_ = LocalSubProcessPep517Executor(
                colored=self.options.is_colored,
                cmd=self.backend_cmd,
                env=self._environment_variables,
                cwd=self._root,
            )

        return self._backend_executor_

    @property
    def backend_cmd(self) -> Sequence[str]:
        return ["python"] + self.backend_args

    @property
    def _environment_variables(self) -> Dict[str, str]:
        env = super()._environment_variables
        backend = os.pathsep.join(str(i) for i in self._backend_paths).strip()
        if backend:
            env["PYTHONPATH"] = backend
        return env

    @contextmanager
    def _send_msg(
        self, cmd: str, result_file: Path, msg: str  # noqa: U100
    ) -> Iterator[ToxCmdStatus]:  # type: ignore[override]
        with self.execute_async(
            cmd=self.backend_cmd,
            cwd=self._root,
            stdin=StdinSource.API,
            show=None,
            run_id=cmd,
            executor=self._backend_executor,
        ) as execute_status:
            execute_status.write_stdin(f"{msg}{os.linesep}")
            yield ToxCmdStatus(execute_status)
        outcome = execute_status.outcome
        if outcome is not None:  # pragma: no branch
            outcome.assert_success()

    def _send(self, cmd: str, **kwargs: Any) -> Tuple[Any, str, str]:
        try:
            if cmd == "prepare_metadata_for_build_wheel":
                # given we'll build a wheel we might skip the prepare step
                if "wheel" in self._builds:
                    result = {
                        "code": 1,
                        "exc_type": "AvoidRedundant",
                        "exc_msg": "will need to build wheel either way, avoid prepare",
                    }
                    raise BackendFailed(result, "", "")
            return super()._send(cmd, **kwargs)
        except BackendFailed as exception:
            raise exception if isinstance(exception, ToxBackendFailed) else ToxBackendFailed(exception) from exception

    def _unexpected_response(self, cmd: str, got: Any, expected_type: Any, out: str, err: str) -> NoReturn:
        try:
            super()._unexpected_response(cmd, got, expected_type, out, err)
        except BackendFailed as exception:
            raise exception if isinstance(exception, ToxBackendFailed) else ToxBackendFailed(exception) from exception

    def requires(self) -> Tuple[Requirement, ...]:
        return self._requires


@impl
def tox_register_tox_env(register: ToxEnvRegister) -> None:
    register.add_package_env(Pep517VirtualEnvPackage)



from pathlib import Path

from tox.config.cli.parser import ToxParser
from tox.config.loader.memory import MemoryLoader
from tox.plugin import impl
from tox.report import HandledError
from tox.session.cmd.run.common import env_run_create_flags
from tox.session.cmd.run.sequential import run_sequential
from tox.session.common import CliEnv, env_list_flag
from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    help_msg = "sets up a development environment at ENVDIR based on the tox configuration specified "
    our = parser.add_command("devenv", ["d"], help_msg, devenv)
    our.add_argument("devenv_path", metavar="path", default=Path("venv").absolute(), nargs="?")
    env_list_flag(our, default=CliEnv("py"), multiple=False)
    env_run_create_flags(our, mode="devenv")


def devenv(state: State) -> int:
    opt = state.options
    opt.skip_missing_interpreters = False
    opt.no_test = False
    opt.package_only = False
    opt.install_pkg = None
    opt.skip_pkg_install = False

    env_list = list(state.env_list(everything=False))
    if len(env_list) != 1:
        raise HandledError(f"exactly one target environment allowed in devenv mode but found {', '.join(env_list)}")
    loader = MemoryLoader(  # these configuration values are loaded from in-memory always (no file conf)
        # dev environments must be of type dev
        usedevelop=True,
        # move it in source
        env_dir=Path(state.options.devenv_path),
    )

    state.options.no_test = True  # do not run the test phase
    state.conf.get_env(env_list[0], loaders=[loader])
    return run_sequential(state)



"""
This package handles provisioning an appropriate tox version per requirements.
"""
import logging
import sys
from argparse import ArgumentParser
from typing import List, Tuple, Union, cast

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version

from tox.config.loader.memory import MemoryLoader
from tox.config.sets import CoreConfigSet
from tox.execute.api import StdinSource
from tox.plugin import impl
from tox.report import HandledError
from tox.session.state import State
from tox.tox_env.errors import Skip
from tox.tox_env.python.pip.req_file import PythonDeps
from tox.tox_env.python.runner import PythonRun
from tox.version import __version__ as current_version

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import PackageNotFoundError, distribution
else:  # pragma: no cover (py38+)
    from importlib_metadata import PackageNotFoundError, distribution  # noqa


@impl
def tox_add_option(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--no-recreate-provision",
        dest="no_recreate_provision",
        help="if recreate is set do not recreate provision tox environment",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--recreate",
        dest="recreate",
        help="recreate the tox environments",
        action="store_true",
    )


@impl
def tox_add_core_config(core: CoreConfigSet) -> None:
    core.add_config(
        keys=["min_version", "minversion"],
        of_type=Version,
        # do not include local version specifier (because it's not allowed in version spec per PEP-440)
        default=Version(current_version.split("+")[0]),
        desc="Define the minimal tox version required to run",
    )
    core.add_config(
        keys="provision_tox_env",
        of_type=str,
        default=".tox",
        desc="Name of the virtual environment used to provision a tox.",
    )

    def add_tox_requires_min_version(requires: List[Requirement]) -> List[Requirement]:  # noqa
        min_version: Version = core["min_version"]
        requires.append(Requirement(f"tox >= {min_version.public}"))
        return requires

    core.add_config(
        keys="requires",
        of_type=List[Requirement],
        default=[],
        desc="Name of the virtual environment used to provision a tox.",
        post_process=add_tox_requires_min_version,
    )


def provision(state: State) -> Union[int, bool]:
    requires: List[Requirement] = state.conf.core["requires"]
    missing: List[Tuple[Requirement, str]] = []
    for package in requires:
        package_name = canonicalize_name(package.name)
        try:
            dist = distribution(package_name)  # type: ignore[no-untyped-call]
        except PackageNotFoundError:
            missing.append((package, "N/A"))
        else:
            if not package.specifier.contains(dist.version, prereleases=True):
                missing.append((package, dist.version))
    if not missing:
        return False
    deps = ", ".join(f"{p} ({ver})" for p, ver in missing)
    msg = "will run in automatically provisioned tox, host %s is missing [requires (has)]: %s"
    logging.warning(msg, sys.executable, deps)
    return run_provision(requires, state)


def run_provision(deps: List[Requirement], state: State) -> int:  # noqa
    """ """
    loader = MemoryLoader(  # these configuration values are loaded from in-memory always (no file conf)
        base=[],  # disable inheritance for provision environments
        package="skip",  # no packaging for this please
        # use our own dependency specification
        deps=PythonDeps("\n".join(str(d) for d in deps), root=state.conf.core["tox_root"]),
        pass_env=["*"],  # do not filter environment variables, will be handled by provisioned tox
        recreate=state.options.recreate and not state.options.no_recreate_provision,
    )
    provision_tox_env: str = state.conf.core["provision_tox_env"]
    state.conf.get_env(provision_tox_env, loaders=[loader])
    tox_env = cast(PythonRun, state.tox_env(provision_tox_env))
    env_python = tox_env.env_python()
    logging.info("will run in a automatically provisioned python environment under %s", env_python)
    try:
        tox_env.setup()
    except Skip as exception:
        raise HandledError(f"cannot provision tox environment {tox_env.conf['env_name']} because {exception}")
    args: List[str] = [str(env_python), "-m", "tox"]
    args.extend(state.args)
    outcome = tox_env.execute(cmd=args, stdin=StdinSource.user_only(), show=True, run_id="provision")
    return cast(int, outcome.exit_code)



from argparse import ArgumentParser
from typing import Any, Callable, TypeVar, cast

import pluggy

from tox.config.main import Config
from tox.config.sets import ConfigSet
from tox.tox_env.register import ToxEnvRegister

from . import NAME

_F = TypeVar("_F", bound=Callable[..., Any])
_spec_marker = pluggy.HookspecMarker(NAME)


def _spec(func: _F) -> _F:
    return cast(_F, _spec_marker(func))


@_spec
def tox_register_tox_env(register: ToxEnvRegister) -> None:  # noqa: U100
    """
    Register new tox environment type. You can register:

    - **run environment**: by default this is a local subprocess backed virtualenv Python
    - **packaging environment**: by default this is a PEP-517 compliant local subprocess backed virtualenv Python

    :param register: a object that can be used to register new tox environment types
    """


@_spec
def tox_add_option(parser: ArgumentParser) -> None:  # noqa: U100
    """
    Add a command line argument. This is the first hook to be called, right after the logging setup and config source
    discovery.

    :param parser: the command line parser
    """


@_spec
def tox_add_core_config(core: ConfigSet) -> None:  # noqa: U100
    """
    Define a new core (non test environment bound) settings for tox. Called the first time the core configuration is
    used (at the start of the provision check).

    :param core: the core configuration object
    """


@_spec
def tox_configure(config: Config) -> None:  # noqa: U100
    """
    Called after command line options are parsed and ini-file has been read.

    :param config: the configuration object
    """


__all__ = (
    "NAME",
    "tox_register_tox_env",
    "tox_add_option",
    "tox_add_core_config",
    "tox_configure",
)



"""
Declare the abstract base class for tox environments that handle the Python language.
"""
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, cast

from packaging.tags import INTERPRETER_SHORT_NAMES
from virtualenv.discovery.py_spec import PythonSpec

from tox.config.main import Config
from tox.tox_env.api import ToxEnv, ToxEnvCreateArgs
from tox.tox_env.errors import Fail, Recreate, Skip


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


class PythonInfo(NamedTuple):
    implementation: str
    version_info: VersionInfo
    version: str
    is_64: bool
    platform: str
    extra: Dict[str, Any]

    @property
    def version_no_dot(self) -> str:
        return f"{self.version_info.major}{self.version_info.minor}"

    @property
    def impl_lower(self) -> str:
        return self.implementation.lower()


class Python(ToxEnv, ABC):
    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        self._base_python: Optional[PythonInfo] = None
        self._base_python_searched: bool = False
        super().__init__(create_args)

    def register_config(self) -> None:
        super().register_config()

        def validate_base_python(value: List[str]) -> List[str]:
            return self._validate_base_python(self.name, value, self.core["ignore_base_python_conflict"])

        self.conf.add_config(
            keys=["base_python", "basepython"],
            of_type=List[str],
            default=self.default_base_python,
            desc="environment identifier for python, first one found wins",
            post_process=validate_base_python,
        )
        self.core.add_config(
            keys=["ignore_base_python_conflict", "ignore_basepython_conflict"],
            of_type=bool,
            default=False,
            desc="do not raise error if the environment name conflicts with base python",
        )
        self.conf.add_constant(
            keys=["env_site_packages_dir", "envsitepackagesdir"],
            desc="the python environments site package",
            value=lambda: self.env_site_package_dir(),
        )
        self.conf.add_constant(
            keys=["env_bin_dir", "envbindir"],
            desc="the python environments binary folder",
            value=lambda: self.env_bin_dir(),
        )
        self.conf.add_constant(
            ["env_python", "envpython"],
            desc="python executable from within the tox environment",
            value=lambda: self.env_python(),
        )

    def _default_pass_env(self) -> List[str]:
        env = super()._default_pass_env()
        if sys.platform == "win32":  # pragma: win32 cover
            env.extend(
                [
                    "SYSTEMDRIVE",
                    "SYSTEMROOT",  # needed for python's crypto module
                    "COMSPEC",  # needed for distutils cygwin compiler
                    "PROCESSOR_ARCHITECTURE",  # platform.machine()
                ]
            )
        env.extend(["REQUESTS_CA_BUNDLE"])
        return env

    def default_base_python(self, conf: "Config", env_name: Optional[str]) -> List[str]:  # noqa
        base_python = None if env_name is None else self.extract_base_python(env_name)
        return [sys.executable if base_python is None else base_python]

    @staticmethod
    def extract_base_python(env_name: str) -> Optional[str]:
        candidates: List[str] = []
        for factor in env_name.split("-"):
            spec = PythonSpec.from_string_spec(factor)
            if spec.implementation is not None:
                if spec.implementation.lower() in INTERPRETER_SHORT_NAMES and env_name is not None:
                    candidates.append(factor)
        if candidates:
            if len(candidates) > 1:
                raise ValueError(f"conflicting factors {', '.join(candidates)} in {env_name}")
            return next(iter(candidates))
        return None

    @staticmethod
    def _validate_base_python(env_name: str, base_pythons: List[str], ignore_base_python_conflict: bool) -> List[str]:
        elements = {env_name}  # match with full env-name
        elements.update(env_name.split("-"))  # and also any factor
        for candidate in elements:
            spec_name = PythonSpec.from_string_spec(candidate)
            if spec_name.implementation is not None and spec_name.implementation.lower() in ("pypy", "cpython"):
                for base_python in base_pythons:
                    spec_base = PythonSpec.from_string_spec(base_python)
                    if any(
                        getattr(spec_base, key) != getattr(spec_name, key)
                        for key in ("implementation", "major", "minor", "micro", "architecture")
                        if getattr(spec_base, key) is not None and getattr(spec_name, key) is not None
                    ):
                        msg = f"env name {env_name} conflicting with base python {base_python}"
                        if ignore_base_python_conflict:
                            return [env_name]  # ignore the base python settings
                        raise Fail(msg)
        return base_pythons

    @abstractmethod
    def env_site_package_dir(self) -> Path:
        """
        If we have the python we just need to look at the last path under prefix.
        E.g., Debian derivatives change the site-packages to dist-packages, so we need to fix it for site-packages.
        """
        raise NotImplementedError

    @abstractmethod
    def env_python(self) -> Path:
        """The python executable within the tox environment"""
        raise NotImplementedError

    @abstractmethod
    def env_bin_dir(self) -> Path:
        """The binary folder within the tox environment"""
        raise NotImplementedError

    def _setup_env(self) -> None:
        """setup a virtual python environment"""
        super()._setup_env()
        self.ensure_python_env()
        self._paths = self.prepend_env_var_path()  # now that the environment exist we can add them to the path

    def ensure_python_env(self) -> None:
        conf = self.python_cache()
        with self.cache.compare(conf, Python.__name__) as (eq, old):
            if old is None:  # does not exist -> create
                self.create_python_env()
            elif eq is False:  # pragma: no branch # exists but changed -> recreate
                raise Recreate(self._diff_msg(conf, old))

    @staticmethod
    def _diff_msg(conf: Dict[str, Any], old: Dict[str, Any]) -> str:
        result: List[str] = []
        added = [f"{k}={v!r}" for k, v in conf.items() if k not in old]
        if added:  # pragma: no branch
            result.append(f"added {' | '.join(added)}")
        removed = [f"{k}={v!r}" for k, v in old.items() if k not in conf]
        if removed:
            result.append(f"removed {' | '.join(removed)}")
        changed = [f"{k}={v!r}->{old[k]!r}" for k, v in conf.items() if k in old and v != old[k]]
        if changed:
            result.append(f"changed {' | '.join(changed)}")
        return f'python {", ".join(result)}'

    @abstractmethod
    def prepend_env_var_path(self) -> List[Path]:
        raise NotImplementedError

    def _done_with_setup(self) -> None:
        """called when setup is done"""
        super()._done_with_setup()
        if self.journal:
            outcome = self.installer.installed()
            self.journal["installed_packages"] = outcome

    def python_cache(self) -> Dict[str, Any]:
        return {
            "version_info": list(self.base_python.version_info),
        }

    @property
    def base_python(self) -> PythonInfo:
        """Resolve base python"""
        if self._base_python_searched is False:
            base_pythons: List[str] = self.conf["base_python"]
            self._base_python_searched = True
            self._base_python = self._get_python(base_pythons)
            if self._base_python is None:
                if self.core["skip_missing_interpreters"]:
                    raise Skip(f"could not find python interpreter with spec(s): {', '.join(base_pythons)}")
                raise NoInterpreter(base_pythons)
            if self.journal:
                value = self._get_env_journal_python()
                self.journal["python"] = value
        return cast(PythonInfo, self._base_python)

    def _get_env_journal_python(self) -> Dict[str, Any]:
        assert self._base_python is not None
        return {
            "implementation": self._base_python.implementation,
            "version_info": tuple(self.base_python.version_info),
            "version": self._base_python.version,
            "is_64": self._base_python.is_64,
            "sysplatform": self._base_python.platform,
            "extra_version_info": None,
        }

    @abstractmethod
    def _get_python(self, base_python: List[str]) -> Optional[PythonInfo]:  # noqa: U100
        raise NotImplementedError

    @abstractmethod
    def create_python_env(self) -> None:
        raise NotImplementedError


class NoInterpreter(Fail):
    """could not find interpreter"""

    def __init__(self, base_pythons: List[str]) -> None:
        self.base_pythons = base_pythons

    def __str__(self) -> str:
        return f"could not find python interpreter matching any of the specs {', '.join(self.base_pythons)}"



import sys
from itertools import zip_longest
from pathlib import Path
from textwrap import dedent

import pytest
from packaging.requirements import Requirement

from tox.pytest import TempPathFactory, ToxProjectCreator
from tox.tox_env.python.virtual_env.package.api import Pep517VirtualEnvPackage
from tox.util.pep517.via_fresh_subprocess import SubprocessFrontend

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution, PathDistribution  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution, PathDistribution  # noqa


@pytest.mark.parametrize(
    "pkg_type",
    ["dev-legacy", "sdist", "wheel"],
)
def test_tox_ini_package_type_valid(tox_project: ToxProjectCreator, pkg_type: str) -> None:
    proj = tox_project({"tox.ini": f"[testenv]\npackage={pkg_type}"})
    result = proj.run("c", "-k", "package_tox_env_type")
    result.assert_success()
    res = result.env_conf("py")["package"]
    assert res == pkg_type
    got_type = result.env_conf("py")["package_tox_env_type"]
    assert got_type == "virtualenv-pep-517"


def test_tox_ini_package_type_invalid(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=bad"})
    result = proj.run("c", "-k", "package_tox_env_type")
    result.assert_failed()
    assert " invalid package config type bad requested, must be one of wheel, sdist, dev-legacy, skip" in result.out


@pytest.fixture(scope="session")
def pkg_with_extras_project(tmp_path_factory: TempPathFactory) -> Path:
    py_ver = ".".join(str(i) for i in sys.version_info[0:2])
    setup_cfg = f"""
    [metadata]
    name = demo
    [options]
    packages = find:
    install_requires =
        platformdirs>=2.1
        colorama>=0.4.3

    [options.extras_require]
    testing =
        covdefaults>=1.2; python_version == '2.7' or python_version == '{py_ver}'
        pytest>=5.4.1; python_version == '{py_ver}'
    docs =
        sphinx>=3
        sphinx-rtd-theme>=0.4.3,<1
    format =
        black>=3
        flake8
    """
    tmp_path = tmp_path_factory.mktemp("prj")
    (tmp_path / "setup.cfg").write_text(dedent(setup_cfg))
    (tmp_path / "setup.py").write_text("from setuptools import setup; setup()")
    toml = '[build-system]\nrequires=["setuptools", "wheel"]\nbuild-backend = "setuptools.build_meta"'
    (tmp_path / "pyproject.toml").write_text(toml)
    return tmp_path


@pytest.fixture(scope="session")
def pkg_with_extras(pkg_with_extras_project: Path) -> PathDistribution:  # type: ignore[no-any-unimported]
    frontend = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(pkg_with_extras_project)[:-1])
    meta = pkg_with_extras_project / "meta"
    result = frontend.prepare_metadata_for_build_wheel(meta)
    return Distribution.at(result.metadata)


def test_load_dependency_no_extra(pkg_with_extras: PathDistribution) -> None:  # type: ignore[no-any-unimported]
    result = Pep517VirtualEnvPackage._dependencies_with_extras(
        [Requirement(i) for i in pkg_with_extras.requires], set()
    )
    for left, right in zip_longest(result, (Requirement("platformdirs>=2.1"), Requirement("colorama>=0.4.3"))):
        assert isinstance(right, Requirement)
        assert str(left) == str(right)


def test_load_dependency_many_extra(pkg_with_extras: PathDistribution) -> None:  # type: ignore[no-any-unimported]
    py_ver = ".".join(str(i) for i in sys.version_info[0:2])
    result = Pep517VirtualEnvPackage._dependencies_with_extras(
        [Requirement(i) for i in pkg_with_extras.requires], {"docs", "testing"}
    )
    exp = [
        Requirement("platformdirs>=2.1"),
        Requirement("colorama>=0.4.3"),
        Requirement("sphinx>=3"),
        Requirement("sphinx-rtd-theme<1,>=0.4.3"),
        Requirement(f'covdefaults>=1.2; python_version == "2.7" or python_version == "{py_ver}"'),
        Requirement(f'pytest>=5.4.1; python_version == "{py_ver}"'),
    ]
    for left, right in zip_longest(result, exp):
        assert isinstance(right, Requirement)
        assert str(left) == str(right)


def test_get_package_deps_different_extras(pkg_with_extras_project: Path, tox_project: ToxProjectCreator) -> None:
    ini = "[testenv:a]\npackage=dev-legacy\nextras=docs\n[testenv:b]\npackage=sdist\nextras=format"
    proj = tox_project({"tox.ini": ini})
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--root", str(pkg_with_extras_project), "-e", "a,b")
    result.assert_success()
    installs = {
        i[0][0].conf.name: i[0][3].cmd[5:]
        for i in execute_calls.call_args_list
        if i[0][3].run_id.startswith("install_package_deps")
    }
    assert installs == {
        "a": ["colorama>=0.4.3", "platformdirs>=2.1", "setuptools", "sphinx-rtd-theme<1,>=0.4.3", "sphinx>=3", "wheel"],
        "b": ["black>=3", "colorama>=0.4.3", "flake8", "platformdirs>=2.1"],
    }


def test_package_root_via_root(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = f"[tox]\npackage_root={demo_pkg_inline}\n[testenv]\npackage=wheel\nwheel_build_env=.pkg"
    proj = tox_project({"tox.ini": ini})
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()


def test_package_root_via_testenv(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = f"[testenv]\npackage=wheel\nwheel_build_env=.pkg\npackage_root={demo_pkg_inline}"
    proj = tox_project({"tox.ini": ini})
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()



import pytest

from tox.pytest import ToxProjectCreator


@pytest.mark.plugin_test()
def test_inline_tox_py(tox_project: ToxProjectCreator) -> None:
    ini = """
    from tox.plugin import impl
    @impl
    def tox_add_option(parser):
        parser.add_argument("--magic", action="store_true")
    """
    project = tox_project({"toxfile.py": ini})
    result = project.run("-h")
    result.assert_success()
    assert "--magic" in result.out



"""
Abstract base API for executing commands within tox environments.
"""
import logging
import sys
import time
from abc import ABC, abstractmethod
from contextlib import contextmanager
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, NoReturn, Optional, Sequence, Tuple, Type, cast

from colorama import Fore

from tox.report import OutErr

from .request import ExecuteRequest, StdinSource
from .stream import SyncWrite

if TYPE_CHECKING:
    from tox.tox_env.api import ToxEnv

ContentHandler = Callable[[bytes], None]
Executor = Callable[[ExecuteRequest, ContentHandler, ContentHandler], int]
LOGGER = logging.getLogger(__name__)


class ExecuteOptions:
    def __init__(self, env: "ToxEnv") -> None:
        self._env = env

    @classmethod
    def register_conf(cls, env: "ToxEnv") -> None:  # noqa
        env.conf.add_config(
            keys=["suicide_timeout"],
            desc="timeout to allow process to exit before sending SIGINT",
            of_type=float,
            default=0.0,
        )
        env.conf.add_config(
            keys=["interrupt_timeout"],
            desc="timeout before sending SIGTERM after SIGINT",
            of_type=float,
            default=0.3,
        )
        env.conf.add_config(
            keys=["terminate_timeout"],
            desc="timeout before sending SIGKILL after SIGTERM",
            of_type=float,
            default=0.2,
        )

    @property
    def suicide_timeout(self) -> float:
        return cast(float, self._env.conf["suicide_timeout"])

    @property
    def interrupt_timeout(self) -> float:
        return cast(float, self._env.conf["interrupt_timeout"])

    @property
    def terminate_timeout(self) -> float:
        return cast(float, self._env.conf["terminate_timeout"])


class ExecuteStatus(ABC):
    def __init__(self, options: ExecuteOptions, out: SyncWrite, err: SyncWrite) -> None:
        self.outcome: Optional[Outcome] = None
        self.options = options
        self._out = out
        self._err = err

    @property
    @abstractmethod
    def exit_code(self) -> Optional[int]:
        raise NotImplementedError

    @abstractmethod
    def wait(self, timeout: Optional[float] = None) -> Optional[int]:  # noqa: U100
        raise NotImplementedError

    @abstractmethod
    def write_stdin(self, content: str) -> None:  # noqa: U100
        raise NotImplementedError

    @abstractmethod
    def interrupt(self) -> None:
        raise NotImplementedError

    def set_out_err(self, out: SyncWrite, err: SyncWrite) -> Tuple[SyncWrite, SyncWrite]:
        res = self._out, self._err
        self._out, self._err = out, err
        return res

    @property
    def out(self) -> bytearray:
        return self._out.content

    @property
    def err(self) -> bytearray:
        return self._err.content

    @property
    def metadata(self) -> Dict[str, Any]:
        return {}


class Execute(ABC):
    """Abstract API for execution of a tox environment"""

    _option_class: Type[ExecuteOptions] = ExecuteOptions

    def __init__(self, colored: bool) -> None:
        self._colored = colored

    @contextmanager
    def call(self, request: ExecuteRequest, show: bool, out_err: OutErr, env: "ToxEnv") -> Iterator[ExecuteStatus]:
        start = time.monotonic()
        try:
            # collector is what forwards the content from the file streams to the standard streams
            out, err = out_err[0].buffer, out_err[1].buffer
            out_sync = SyncWrite(out.name, out if show else None)
            err_sync = SyncWrite(err.name, err if show else None, Fore.RED if self._colored else None)
            with out_sync, err_sync:
                instance = self.build_instance(request, self._option_class(env), out_sync, err_sync)
                with instance as status:
                    yield status
                exit_code = status.exit_code
        finally:
            end = time.monotonic()
        status.outcome = Outcome(
            request, show, exit_code, out_sync.text, err_sync.text, start, end, instance.cmd, status.metadata
        )

    @abstractmethod
    def build_instance(
        self, request: ExecuteRequest, options: ExecuteOptions, out: SyncWrite, err: SyncWrite  # noqa: U100
    ) -> "ExecuteInstance":
        raise NotImplementedError

    @classmethod
    def register_conf(cls, env: "ToxEnv") -> None:
        cls._option_class.register_conf(env)


class ExecuteInstance(ABC):
    """An instance of a command execution"""

    def __init__(self, request: ExecuteRequest, options: ExecuteOptions, out: SyncWrite, err: SyncWrite) -> None:
        self.request = request
        self.options = options
        self._out = out
        self._err = err

    @property
    def out_handler(self) -> ContentHandler:
        return self._out.handler

    @property
    def err_handler(self) -> ContentHandler:
        return self._err.handler

    @abstractmethod
    def __enter__(self) -> ExecuteStatus:
        raise NotImplementedError

    @abstractmethod
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],  # noqa: U100
        exc_val: Optional[BaseException],  # noqa: U100
        exc_tb: Optional[TracebackType],  # noqa: U100
    ) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def cmd(self) -> Sequence[str]:
        raise NotImplementedError


class Outcome:
    """Result of a command execution"""

    OK = 0

    def __init__(
        self,
        request: ExecuteRequest,
        show_on_standard: bool,
        exit_code: Optional[int],
        out: str,
        err: str,
        start: float,
        end: float,
        cmd: Sequence[str],
        metadata: Dict[str, Any],
    ):
        """
        Create a new execution outcome.

        :param request: the execution request
        :param show_on_standard: a flag indicating if the execution was shown on stdout/stderr
        :param exit_code: the exit code for the execution
        :param out: the standard output of the execution
        :param err: the standard error of the execution
        :param start: a timer sample for the start of the execution
        :param end: a timer sample for the end of the execution
        :param cmd: the command as executed
        :param metadata: additional metadata attached to the execution
        """
        self.request = request  #: the execution request
        self.show_on_standard = show_on_standard  #: a flag indicating if the execution was shown on stdout/stderr
        self.exit_code = exit_code  #: the exit code for the execution
        self.out = out  #: the standard output of the execution
        self.err = err  #: the standard error of the execution
        self.start = start  #: a timer sample for the start of the execution
        self.end = end  #: a timer sample for the end of the execution
        self.cmd = cmd  #: the command as executed
        self.metadata = metadata  #: additional metadata attached to the execution

    def __bool__(self) -> bool:
        return self.exit_code == self.OK

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: exit {self.exit_code} in {self.elapsed:.2f} seconds"
            f" for {self.request.shell_cmd}"
        )

    def assert_success(self) -> None:
        """Assert that the execution succeeded"""
        if self.exit_code is not None and self.exit_code != self.OK:
            self._assert_fail()
        self.log_run_done(logging.INFO)

    def _assert_fail(self) -> NoReturn:
        if self.show_on_standard is False:
            if self.out:
                sys.stdout.write(self.out)
                if not self.out.endswith("\n"):
                    sys.stdout.write("\n")
            if self.err:
                sys.stderr.write(Fore.RED)
                sys.stderr.write(self.err)
                sys.stderr.write(Fore.RESET)
                if not self.err.endswith("\n"):
                    sys.stderr.write("\n")
        self.log_run_done(logging.CRITICAL)
        raise SystemExit(self.exit_code)

    def log_run_done(self, lvl: int) -> None:
        """
        Log that the run was done.

        :param lvl: the level on what to log as interpreted by :func:`logging.log`
        """
        req = self.request
        metadata = ""
        if self.metadata:
            metadata = f" {', '.join(f'{k}={v}' for k, v in self.metadata.items())}"
        LOGGER.log(
            lvl,
            "exit %s (%.2f seconds) %s> %s%s",
            self.exit_code,
            self.elapsed,
            req.cwd,
            req.shell_cmd,
            metadata,
        )

    @property
    def elapsed(self) -> float:
        """:return: time the execution took in seconds"""
        return self.end - self.start

    def out_err(self) -> Tuple[str, str]:
        """:return: a tuple of the standard output and standard error"""
        return self.out, self.err


__all__ = (
    "ContentHandler",
    "Outcome",
    "Execute",
    "ExecuteInstance",
    "ExecuteOptions",
    "ExecuteStatus",
    "StdinSource",
)
